==> dapu-0.0.2-py3-none-any.whl: dapu/enabler.py <==
from dapu.process import DapuProcess
from dapu.perks import halt
from dapu.util_yaml import load_yaml
from dapu.process import logging
from typing import Callable
from dapu.perks import reconf_logging

def interpret_string_as_yaml(yaml_content_string: str) -> dict | None:
    import yaml
    if yaml_content_string is None:
        return None
    return yaml.load(yaml_content_string, Loader = yaml.Loader) or None


def read_included_file(file_name: str) -> str | None:
    """
    Return content on one resource file included into package ver submodule
    """
    module_name = 'dapu.ver'

    from importlib.resources import files # needs 3.7+, actually 3.9+
    
    try:
        # Open the file as a resource within the package
        # ver dir must have (empty) __init__.py file inside
        # and ver dir must be built in during package build
        # and to run it without build and local install, let ver be subdir with __init__ 
        # old way, deprecated in 3.11 (newest ver is 3.13):        
        # with importlib.resources.open_text(module_name, file_name, encoding="UTF-8") as file:
        # new way, needs py 3.9, encoding="utf-8" is default
        with files(module_name).joinpath(file_name).open('r') as file: # 3.9+
            content = file.read()
        return content
    
    except Exception as e1:
        logging.error(f"Resource file {file_name} does not exist: {e1}")
        return None


class DapuEnabler(DapuProcess):
    """
    Does versioning of meta tables (registry, agenda etc) in meta schema
    Run it on after every deploy/ugrade of package dbpuller to assure thet meta structure corresponds to current version of code  
    """

    def run(self):
        reconf_logging(self.context.log_level) # with some reason init-time is lost somewhere (datacontroller??)
        logging.info(f"Working directory is {self.context.work_dir}")
        
        if (start_time := self.validate_connection()) is None:
            logging.error(f"Failed to connect to database")

        logging.debug(f"Database time is '{start_time}'")
        
        # build-in (core) versioning
        core_migra_index: dict = self.read_included_toc()
        core_access_function: Callable = read_included_file # fn which can read file content from module files
        self.run_migration(core_migra_index, core_access_function) # on error halts
        logging.info(f"Core versioning done")
        
        # work_dir offered (custom) versioning
        custom_migra_index: dict = self.read_custom_ver_toc()
        custom_access_function: Callable = self.read_custom_file # fn which can read file from local file system
        self.run_migration(custom_migra_index, custom_access_function)
        logging.info(f"Project versioning done")
        self.disconnect_main()


    def validate_connection(self) -> str | None:
        """
        Lets ask current time from database for realistic connection validation (None is failure)
        """
        return self.get_database_time()
    

    def read_included_toc(self) -> dict | None:
        """
        Metatables version upgrade files (sql) must reside in submodule "ver"
        index file is changes.yaml
        index file points in correct order to SQL-files
        index file have declared alias (so same metatable can be used for custom versioning too) 
        """
        return interpret_string_as_yaml(read_included_file(self.context.ENABLER_INDEX_FILE_NAME))
    
    
    def read_custom_ver_toc(self) -> dict | None:
        """
        Read from self.work_dir subfolder "ver"
        """
        dir_ver = self.context.full_name_from_root(self.context.ENABLER_SUB_DIR_NAME)
        logging.info(f"Custom versioning dir is {dir_ver}")
        return load_yaml(dir_ver, self.context.ENABLER_INDEX_FILE_NAME, empty_as={})


    def read_custom_file(self, file_name: str) -> dict | None:
        with open(self.context.full_name_from_ver(file_name), 'r', encoding='utf-8') as file:
            content = file.read()
        return content
    

    def apply_ver(self, sql: str, folder_alias: str, file_name: str, remarks: str='') -> bool:
        """
        Applies internal (package) versioning SQL and saves/logs it
        If dry_run, then just prints out (and all if true, but no ver upgrade marks) 
        """
        if self.apply_sql(sql, with_replacements=True, with_commit=True): # MAIN WORK
            version_table = self.context.find_registry_table_full_name("version")
            file_name = file_name.replace("'", "")
            sql = f"""INSERT INTO {version_table} (folder_alias, file_name, remarks) 
                VALUES ('{folder_alias}', '{file_name}', '{remarks}')"""
            if self.apply_sql(sql, with_replacements=False, with_commit=True): # registering what was done
                logging.info(f"{sql}")
                logging.info(f"{file_name} applied")
                return True
            else:
                logging.info(f"{file_name} WAS OK, but after-actions NOT, rollbacking")
                return False
        else:
            logging.error(f"{file_name} NOT applied")
            return False
    
    
    def apply_sql(self, sql:str, with_replacements:bool=True, with_commit:bool=False) -> bool:
        """
        Runs one SQL file (may have many DDL commands)
        """
        if with_replacements:
            sql = self.context.replace_compatibility(sql)
        try:
            # apply, but not commit (commit will be after version is saved too)
            self.context.target(sql, False, on_success_commit=with_commit)
            print(sql)
            print(f"{with_commit=}")
            print(self.context.list_of_profiles)
            return True
        except Exception as e1:
            logging.debug(sql)
            logging.error(str(e1))
            return False


    def get_applied_versions(self, alias) -> list[str]:
        """
        Get short file names (for given alias) what are applied, order is not important
        """
        version_table = self.context.find_registry_table_full_name("version")
        sql = f"""SELECT file_name
            FROM {version_table}
            WHERE folder_alias = '{alias}'
            """
        try:
            return [rec[0] for rec in self.context.target(sql)]
        except Exception as e1: # ignore is needed for very first run
            logging.debug(str(e1)) # we are going to ignore it, but let output it anyway
            return []
    
    
    def run_migration(self, migra_index: dict, method_for_files: Callable) -> None:
        """
        Runs one migration line, either build-in (core) then files must accessed using importlib
        or custom from ver subfolder using regular file reading
        """
        if migra_index is None:
            halt(91, "Nothing to do. Missing index file. This cannot be true")
        alias = migra_index.get('alias', '').replace("'", "").strip()
        if alias is None:
            halt(92, "Missing alias in index file. Big mess may happen")
        applied_files = self.get_applied_versions(alias)
        file_names = migra_index.get('files')
        if not file_names:
            halt(93, f"No files referenced from index file for alias '{alias}'. Unbelievable!")
        
        for pos, step_info in enumerate(file_names, 1):
            if step_info.get('file') is None:
                continue # nothing to apply
            file_name = step_info.get('file')
            if file_name in applied_files: 
                continue # already applied
            logging.debug(f"Migration for alias {alias} step {pos} file {file_name}")
            sql = method_for_files(file_name) # here is argument of function which is function
            remarks = step_info.get('remarks', '').replace("'", "").strip()
            if not self.apply_ver(sql, alias, file_name, remarks):
                halt(94, f"step {pos}, error on sql in file {file_name}")
        logging.debug(f"Migration of {alias} is done")
        return None # if any error, then global HALT, so after failing in core, custom cannot run

==> dapu-0.0.4-py3-none-any.whl: dapu/enabler.py <==
from dapu.process import DapuProcess
from dapu.process import logging
from typing import Callable
from dapu.perks import halt, reconf_logging, read_package_file, interpret_string_as_yaml



class DapuEnabler(DapuProcess):
    """
    Does versioning of meta tables (registry, agenda etc) in meta schema
    Run it on after every deploy/ugrade of package dbpuller to assure thet meta structure corresponds to current version of code  
    """

    def run(self):
        reconf_logging(self.context.LOG_LEVEL) # with some reason init-time is lost somewhere (datacontroller??)
        logging.info(f"Working directory is {self.context.work_dir}")
        
        if (start_time := self.validate_connection()) is None:
            logging.error(f"Failed to connect to database")

        logging.debug(f"Database time is '{start_time}'")
        
        # build-in (core) versioning
        core_migra_index: dict = self.read_included_toc()
        core_access_function: Callable = self.read_included_file # fn which can read file content from module files
        self.run_migration(core_migra_index, core_access_function) # on error halts
        logging.info(f"Core versioning done")
        
        # work_dir offered (custom) versioning
        custom_migra_index: dict = self.read_custom_ver_toc()
        custom_access_function: Callable = self.read_custom_file # fn which can read file from local file system
        self.run_migration(custom_migra_index, custom_access_function)
        logging.info(f"Project versioning done")
        self.disconnect_main()


    def read_included_file(self, file_name: str) -> str | None:
        """
        Wrapper who knows (how to find) versioning module name. And then calls read_package_file.
        Reads file content using importlib (so actual place of file is unknown)
        """
        versioning_scripts_package = self.context.PACKAGE_NAME + '.' + self.context.PACKAGE_VER_SUB_DIR # 'dapu.ver'
        return read_package_file(versioning_scripts_package, file_name)


    def read_custom_file(self, file_name: str) -> str | None:
        """
        Read file content regular way from file system, knowing that context holds info to file path for versioning
        """
        with open(self.context.full_name_from_ver(file_name), 'r', encoding='utf-8') as file:
            content = file.read()
        return content
    

    def read_included_toc(self) -> dict | None:
        """
        Metatables version upgrade files (sql) must reside in submodule "ver" (self.context.PACKAGE_VER_SUB_DIR)
        Index file is changes.yaml.  
        Index file points in correct order to SQL-files.  
        Index file have declared alias (so same metatable can be used for custom versioning too) 
        """
        versioning_scripts_package = self.context.PACKAGE_NAME + '.' + self.context.PACKAGE_VER_SUB_DIR # 'dapu.ver'
        return interpret_string_as_yaml(read_package_file(versioning_scripts_package, self.context.CHANGES_INDEX_FILE_NAME))
    
    
    def read_custom_ver_toc(self) -> dict | None:
        """
        Read from self.work_dir subfolder "ver"
        """
        yaml_string = self.read_custom_file(self.context.CHANGES_INDEX_FILE_NAME)
        return interpret_string_as_yaml(yaml_string)


    def apply_ver(self, sql: str, folder_alias: str, file_name: str, remarks: str='') -> bool:
        """
        Applies internal (package) versioning SQL and saves/logs it
        If dry_run, then just prints out (and all if true, but no ver upgrade marks) 
        """
        if self.apply_sql(sql, with_replacements=True, with_commit=True): # MAIN WORK
            version_table = self.context.find_registry_table_full_name("version")
            file_name = file_name.replace("'", "")
            sql = f"""INSERT INTO {version_table} (folder_alias, file_name, remarks) 
                VALUES ('{folder_alias}', '{file_name}', '{remarks}')"""
            if self.apply_sql(sql, with_replacements=False, with_commit=True): # registering what was done
                logging.info(f"{sql}")
                logging.info(f"{file_name} applied")
                return True
            else:
                logging.info(f"{file_name} WAS OK, but after-actions NOT, rollbacking")
                return False
        else:
            logging.error(f"{file_name} NOT applied")
            return False
    
    
    def apply_sql(self, sql:str, with_replacements:bool=True, with_commit:bool=False) -> bool:
        """
        Runs one SQL file (may have many DDL commands)
        """
        if with_replacements:
            sql = self.context.replace_compatibility(sql)
        try:
            # apply, but not commit (commit will be after version is saved too)
            self.context.target(sql, False, on_success_commit=with_commit)
            print(sql)
            print(f"{with_commit=}")
            print(self.context.list_of_profiles)
            return True
        except Exception as e1:
            logging.debug(sql)
            logging.error(str(e1))
            return False


    def get_applied_versions(self, alias) -> list[str]:
        """
        Get short file names (for given alias) what are applied, order is not important
        """
        version_table = self.context.find_registry_table_full_name("version")
        sql = f"""SELECT file_name
            FROM {version_table}
            WHERE folder_alias = '{alias}'
            """
        try:
            return [rec[0] for rec in self.context.target(sql)]
        except Exception as e1: # ignore is needed for very first run
            logging.debug(str(e1)) # we are going to ignore it, but let output it anyway
            return []
    
    
    def run_migration(self, migra_index: dict, method_for_files: Callable) -> None:
        """
        Runs one migration line, either build-in (core) then files must accessed using importlib
        or custom from ver subfolder using regular file reading
        """
        if migra_index is None:
            halt(91, "Nothing to do. Missing index file. This cannot be true")
        alias = migra_index.get('alias', '').replace("'", "").strip()
        if not alias: # both None and '' are bad
            halt(92, "Missing alias in index file. Big mess may happen")
        applied_files = self.get_applied_versions(alias)
        file_names = migra_index.get('files')
        if not file_names:
            halt(93, f"No files referenced from index file for alias '{alias}'. Unbelievable!")
        for pos, step_info in enumerate(file_names, 1):
            if step_info.get('file') is None:
                continue # nothing to apply
            file_name = step_info.get('file')
            if file_name in applied_files: 
                continue # already applied
            logging.debug(f"Migration for alias {alias} step {pos} file {file_name}")
            sql = method_for_files(file_name) # here is argument of function which is function
            remarks = step_info.get('remarks', '').replace("'", "").strip()
            if not self.apply_ver(sql, alias, file_name, remarks):
                halt(94, f"step {pos}, error on sql in file {file_name}")
        logging.debug(f"Migration of {alias} is done")
        return None # if any error, then global HALT, so after failing in core, custom cannot run


    def validate_connection(self) -> str | None:
        """
        Asks current time from database = realistic connection validation (None is failure)
        """
        return self.get_database_time()
